/****************************************************************** 
 * Copyright (c) 2003, Mario Hochreiter, Student of 
 * Software Engineering at FH-Hagenberg, Austria. 
 * Project member of the isac team at the Institute for 
 * Software-Technologie, Graz University of Technology, Austria. 
 * 
 * Use is subject to license terms.
 ******************************************************************/

package isac.gui;

import isac.useractions.IUIAction;
import isac.useractions.UserAction;
import isac.util.InformationProcessor;
import isac.browserdialog.BDialog;
import isac.gui.browser.BrowserFrame;
import isac.gui.browser.BrowserFrameRMI;
import isac.gui.browser.panels.ContentPanel;
import isac.gui.browser.panels.ExampleBrowserPanel;
import isac.gui.browser.panels.MethodBrowserPanel;
import isac.gui.browser.panels.ProblemBrowserPanel;
import isac.gui.util.*;
import isac.session.ISessionManager;
import isac.util.Formalization;
import isac.util.parser.XMLExampleParser;
import isac.util.parser.XMLMethodParser;
import isac.util.parser.XMLProblemParser;
import isac.wsdialog.IWorksheetDialog;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;

import javax.swing.*;
import javax.swing.plaf.basic.BasicInternalFrameUI;

import org.apache.log4j.Logger;
import org.w3c.dom.Node;

/**
 * Main window application for the ISAC system. Responsible for showing the user
 * all requested information and for the interaction with the user.
 * 
 * @author Mario Hochreiter
 * @author Manuel Koschuch
 * @version $Revision$
 *  
 */
public class WindowApplication extends JFrame implements ActionListener,
        ILoginListener, IToWorksheetUser {
    /**
     * For serialization purposes
     */
    private static final long serialVersionUID = 1L;

    static Logger logger = Logger.getLogger(WindowApplication.class.getName());

    /** Local settings for the user, WN0505 unused */
    private Locale locale;

    /** Container for the internal frames, which will contain the worksheets */
    private JDesktopPane worksheetDesktopPane;

    private JInternalFrame contentFrame;

    // private JInternalFrame tryMatchFrame;

    // private JInternalFrame tryRefineFrame;

    private JInternalFrame activeFrame;

    /** login dialoge */
    private LoginScreen loginScreen;

    // private CalcHeadPanel tryMatchPanel;

    // private CalcHeadPanel tryRefinePanel;

    private ContentPanel contentPanel;

    /** Top container that contains all elements of this frame */
    private JSplitPane splitPane;

    /** The most recent active worksheet */
    //private Worksheet lastActiveWS;
    /** Required to logon a user and to obtain the <code>WSDialog</code> */
    // private ISessionManager sessionDialog;//WN050415 unused
    /** Unique session_id_ for this WindowApplication */
    private String session_id_;

    /** Required to logon a user and to obtain the <code>WSDialog</code> */
    private ISessionManager session_dialog_;

    // for the constructor of the worksheet
    private final WindowApplication windowApplication_;

    // LK 04.08.30
    // Menu Buttons
    // left side
    private JButton newWorksheetButton;

    private JButton exampleBrowserButton;

    private JButton problemBrowserButton;

    private JButton methodBrowserButton;

    private JButton theoryBrowserButton;

    // right side

    // hash map containing the dynamical created buttons
    private HashMap dynamic_buttons_;

    private HashMap dynamic_actions_;

    // private JButton nextStepButton;

    // private JButton autocalcButton;

    private JButton tacticsButton;

    private BrowserFrame problem_browser_frame_;

    private BrowserFrame example_browser_frame_;

    private BrowserFrame method_browser_frame_;

    /**
     * Panel that displays the different "browseable" hierarchies (e.g. the
     * example Browser).
     */
    // private BrowserPanel browserPanel;
    public WindowApplication() {
        locale = Locale.getDefault();

        this.setSize(new Dimension(1024, 768));
        Point screenCenter = this.screenCenter();
        this.setLocation(screenCenter);

        Container contentPane = this.getContentPane();
        this.buildMenuBar();

        // WN050708 doesnt input here due to SwingUtilities.invokeLater ?...
        // loginScreen = new LoginScreen();
        // loginScreen.addListener(this);
        // loginScreen.show();
        // ... but immediately loadHierarchy via createBrowserPanel()...

        // initialize hashMaps
        dynamic_actions_ = new HashMap();
        dynamic_buttons_ = new HashMap();

        this.createWorksheetPane();
        this.createBrowserFrames();
        this.buildSplitPane();

        contentPane.add(splitPane);

        JFrame.setDefaultLookAndFeelDecorated(true);
        this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.hide();

        loginScreen = new LoginScreen();
        loginScreen.addListener(this);
        loginScreen.show();

        windowApplication_ = this;
    }

    /* LK 04.08.27 */
    /**
     * login method called from <code>LoginScreen</code> Initialize RMI
     * connection to <code>ObjectManager</code> and
     * <code>SessionManager</code>.
     */
    private void login(String username, String password) {
        logger.fatal("WA->SM: login");
        session_id_ = InformationProcessor.login(username, password);
        if (session_id_ != null) {
            session_dialog_ = InformationProcessor.getSDialog();
            loginScreen.dispose();
            this.show();
        } else {
            loginScreen.showErrorMsg();
        }
    }

    /**
     * Create a new WorksheetDialog that can be connected to a new Worksheet
     */
    private IWorksheetDialog newDialog() {
        logger.fatal("WA->SM: newDialog()");
        String dialog_id = null;
        IWorksheetDialog dialog = null;
        try {
            dialog_id = session_dialog_.openWorksheetDialog(session_id_);
            dialog = session_dialog_.getWorksheetDialog(dialog_id);
        } catch (RemoteException e) {
            logger.error("newDialog", e.detail);
        }
        logger.fatal("WA<--DG: newDialog <- dialogId=" + dialog_id);
        return dialog;
    }

    // ---------- Load the different hierarchies---------------------------

    private Hierarchy loadExpHierarchy() {
        String dummy = InformationProcessor.loadHierarchy(session_id_, "exp");
        System.out.println("WN##### loadExpHierarchy, dummy=" + dummy);
        XMLExampleParser expParser = new XMLExampleParser(dummy);
        Node root = expParser.getRoot();
        Hierarchy hierarchy = expParser.createExampleNodesHierarchy(root);
        return hierarchy;
    }

    private Hierarchy loadPblHierarchy() {
        String dummy = InformationProcessor.loadHierarchy(session_id_, "pbl");
        XMLProblemParser pblParser = new XMLProblemParser(dummy);
        Node root = pblParser.getRoot();
        Hierarchy hierarchy = pblParser.createProblemNodesHierarchy(root);
        return hierarchy;
    }

    private Hierarchy loadMetHierarchy() {
        String dummy = InformationProcessor.loadHierarchy(session_id_, "met");
        XMLMethodParser metParser = new XMLMethodParser(dummy);
        Node root = metParser.getRoot();
        Hierarchy hierarchy = metParser.createMethodNodesHierarchy(root);
        return hierarchy;
    }

    private void buildMenuBar() {

        JMenuBar menuBar = new JMenuBar();

        // left side
        newWorksheetButton = new JButton("NEW");
        newWorksheetButton.setName("New");
        newWorksheetButton.addActionListener(this);
        exampleBrowserButton = new JButton("Examples");
        exampleBrowserButton.setName("Examples");
        exampleBrowserButton.addActionListener(this);
        problemBrowserButton = new JButton("Problems");
        problemBrowserButton.setName("Problems");
        problemBrowserButton.addActionListener(this);
        methodBrowserButton = new JButton("Methods");
        methodBrowserButton.setName("Methods");
        methodBrowserButton.addActionListener(this);
        theoryBrowserButton = new JButton("Theories");
        theoryBrowserButton.setName("Theories");
        theoryBrowserButton.addActionListener(this);

        // right side
        // nextStepButton = new JButton("Next");
        // nextStepButton.setName("Next");
        // nextStepButton.addActionListener(this);
        // autocalcButton = new JButton("Auto");
        // autocalcButton.setName("Auto");
        // autocalcButton.addActionListener(this);
        tacticsButton = new JButton("Tactics");
        tacticsButton.setName("Tactics");
        tacticsButton.addActionListener(this);

        // add left menu buttons
        menuBar.add(newWorksheetButton);
        menuBar.add(exampleBrowserButton);
        menuBar.add(problemBrowserButton);
        menuBar.add(methodBrowserButton);
        menuBar.add(theoryBrowserButton);
        // seperate
        menuBar.add(new JLabel("              "));
        menuBar.add(new JLabel("              "));
        menuBar.add(new JLabel("              "));
        menuBar.add(new JLabel("              "));
        menuBar.add(new JLabel("              "));
        menuBar.add(new JLabel("              "));
        // add right menu buttons
        // menuBar.add(nextStepButton);
        // menuBar.add(autocalcButton);
        menuBar.add(tacticsButton);

        this.setJMenuBar(menuBar);
    }

    public void addToMenuBar(IUIAction action) {
        int action_id = action.getAction();
        // check if a button is already in the menu
        if (!dynamic_buttons_.containsKey(new Integer(action_id))) {
            JButton dynamic_button = new JButton(action.getLabel());
            dynamic_button.setName(action.getLabel());
            dynamic_button.addActionListener(this);

            dynamic_buttons_.put(new Integer(action_id), dynamic_button);
            dynamic_actions_.put(dynamic_button, action);
            JMenuBar menuBar = this.getJMenuBar();
            menuBar.add(dynamic_button);
            this.setJMenuBar(menuBar);

        }
    }

    public boolean removeFromMenuBar(IUIAction action) {
        int action_id = action.getAction();
        // does not contain the key that should be removed
        if (!dynamic_buttons_.containsKey(new Integer(action_id))) {
            return false;
        } else {
            JButton button = (JButton) dynamic_buttons_.get(new Integer(
                    action_id));
            dynamic_actions_.remove(button);
            dynamic_buttons_.remove(new Integer(action_id));
            JMenuBar menuBar = this.getJMenuBar();
            menuBar.remove(button);
            this.setJMenuBar(menuBar);
            return true;
        }

    }

    private void createBrowserFrames() {
        example_browser_frame_ = new BrowserFrame("Excample Browser",
                new ExampleBrowserPanel(this, loadExpHierarchy()));
        method_browser_frame_ = new BrowserFrame("Method Browser",
                new MethodBrowserPanel(this, loadMetHierarchy()));
    }

    /**
     * Open a new Worksheet and start the modeling and specification phase.
     * 
     * @param f
     *            Formalization (can be empty see UC 2.5.4) that the math engine
     *            needs to start a calculation
     * @param source
     *            Source of the Formalization (e.g. ExampleHierarchy or from
     *            Scratch)
     * @param type
     *            Requested ViewType for the CalcHead
     * @throws RemoteException
     */
    public void openNewWorksheet(final IToWorksheetUser winAppl,
            final Formalization f, final int source) throws RemoteException {

        // InternalFrame, holding the Frames for the Worksheet and the
        // corresponding CalcHeadPanel
        JInternalFrame internalFrame = createInternalFrame("Worksheet", true,
                true, false, false);
        internalFrame.setResizable(true);

        // DesktopPane, added to internalFrame to be able to hae multiple
        // InternalFrames in this InternalFrame
        final JDesktopPane desktop = new JDesktopPane();
        desktop.setSize(600, 600);
        desktop.setDesktopManager(new DefaultDesktopManager() {
            public void activateFrame(JInternalFrame jif) {
                super.activateFrame(jif);

            }
        });
        desktop.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);

        internalFrame.getContentPane().add(desktop, BorderLayout.CENTER);

        final IWorksheetDialog dialog = newDialog();

        final SwingWorker worker = new SwingWorker() {
            public Object construct() {
                Worksheet ws = new Worksheet(winAppl, locale, dialog, f,
                        source, desktop);
                return ws;
            }

            public void finished() {
                setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
            }
        };
        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        worker.start();

        // new InternalFrame for the Worksheet
        JInternalFrame ws_frame = new JInternalFrame();
        ((BasicInternalFrameUI) ws_frame.getUI()).setNorthPane(null);
        ws_frame.setSize(600, this.getHeight() - 42);
        ws_frame.setVisible(true);
        ws_frame.setResizable(true);

        ws_frame.getContentPane().add((Worksheet) worker.get(),
                BorderLayout.CENTER);
        desktop.add(ws_frame);
        ws_frame.setFocusable(false);

        int width, height = 0;
        height = internalFrame.getPreferredSize().height;
        width = internalFrame.getPreferredSize().width;
        // make Frame a little smaller than the window
        height = this.getHeight() - 42;
        width = width < 600 ? 600 : width;
        internalFrame.setPreferredSize(new Dimension(width, height));
        internalFrame.pack();
        internalFrame.setVisible(true);
        this.addInternalFrame(internalFrame);

    }

    private JInternalFrame createInternalFrame(String title, boolean resizable,
            boolean closable, boolean maximizable, boolean iconifiable) {
        JInternalFrame internalFrame = new JInternalFrame(title, resizable,
                closable, maximizable, iconifiable);
        return internalFrame;
    }

    private void createWorksheetPane() {
        worksheetDesktopPane = new JDesktopPane();
        Dimension d = new Dimension();
        d.setSize(this.getWidth() * 0.60, 0.0);
        worksheetDesktopPane.setMinimumSize(d);
        worksheetDesktopPane.setDesktopManager(new DefaultDesktopManager() {
            public void activateFrame(JInternalFrame jif) {
                super.activateFrame(jif);
                activeFrame = jif;
            }
        });
        // "ghost frame" dragging; very fast but weired for some user
        worksheetDesktopPane.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);
    }

    /**
     * Get the Worksheet on top of the others; this may be below knowledge
     * browsers and other components. Precondition: a Worksheet is a Component
     * in a JInternalFrame in a JDesktopPane.
     * 
     * FIXME.WN050818 search along the gui's tree structure
     * 
     * @return The active Worksheet and null if no Worksheet is open
     */
    public Worksheet getActiveWorksheet() {
        Component comp;
        if (activeFrame != null) {
            JDesktopPane pane = ((JDesktopPane) activeFrame.getContentPane()
                    .getComponent(0));
            for (int i = 0; i < pane.getComponentCount(); i++) {
                try {
                    JInternalFrame frame = (JInternalFrame) pane
                            .getComponent(i);
                    for (int j = 0; j < frame.getComponentCount(); j++) {
                        comp = frame.getContentPane().getComponent(j);
                        if (comp instanceof Worksheet) {
                            return (Worksheet) comp;
                        }
                    }
                } catch (ClassCastException e) {
                }
            }
        }
        //        ((JInternalFrame) ((JDesktopPane) activeFrame.getContentPane()
        //                .getComponent(0)).getComponent(0)).getContentPane()
        //                .getComponent(0);
        return null;
    }

    /**
     * @return the id of the topmost Worksheet or null if there is none.
     */
    public String getActiveWSDID() {
        if (getActiveWorksheet() == null)
            return null;
        else
            return getActiveWorksheet().getWorksheetID();
    }

    private void buildSplitPane() {
        splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, false);
        // split_pane_.setLeftComponent(browserPanel);
        splitPane.setRightComponent(worksheetDesktopPane);
        splitPane.setOneTouchExpandable(true);
        splitPane.setDividerLocation(250);
    }

    private Point screenCenter() {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        int midX = screenSize.width / 2 - this.getWidth() / 2;
        int midY = screenSize.height / 2 - this.getHeight() / 2;
        return new Point(midX, midY);
    }

    private void addInternalFrame(JInternalFrame frame) {
        worksheetDesktopPane.add(frame);
        worksheetDesktopPane.getDesktopManager().activateFrame(frame);
        frame.toFront();
        frame.grabFocus();
    }

    /**
     * Get the session id of the current connection.
     * 
     * @return Systemwide unique session id
     */
    public String getSessionId() {
        return session_id_;
    }

    /**
     * Get the content Panel.
     * 
     * @return ContentPanel
     */
    public ContentPanel getContentPanel() {
        if (contentPanel == null) {
            this.contentPanel = new ContentPanel(session_id_);
        }
        return contentPanel;
    }

    /**
     * Get the currently used locale.
     * 
     * @return currently used locale.
     */
    public Locale getLocale() {
        return locale;
    }

    /**
     * Show the content frame which includes a window that shows the content of
     * a selected entry in the hierarchy_.
     */
    public void showContentFrame() {
        if (contentFrame == null) {
            contentFrame = new JInternalFrame("Content Window", true, true,
                    true);
            this.addInternalFrame(contentFrame);
            contentFrame.getContentPane()
                    .add(contentPanel, BorderLayout.CENTER);
            contentFrame.pack();
            contentFrame.setVisible(true);
        } else if (!contentFrame.isVisible()) {
            this.addInternalFrame(contentFrame);
            contentFrame.setVisible(true);
        }
    }

    // ----------ActionListener----------------------------------------

    public void actionPerformed(ActionEvent e) {
        if (dynamic_buttons_.containsValue(e.getSource())) {
            System.out.println("----------------");
            System.out.println(getActiveWorksheet());
            getActiveWorksheet().notifyUserAction(
                    new UserAction(((IUIAction) dynamic_actions_.get(e
                            .getSource())).getAction()));
        }
        if (e.getSource() == problemBrowserButton) {
            if (problem_browser_frame_ == null) {
                generateProblemBrowser();
                setProblemBrowserVisible();
            } else if (!problem_browser_frame_.isVisible()) {
                setProblemBrowserVisible();
            }
        } else if (e.getSource() == newWorksheetButton) {
            // JOptionPane.showMessageDialog(this, "not yet implemented");
            // WN050504...
            try {
                this.openNewWorksheet(this, new Formalization(),
                        IWorksheetDialog.STARTFROM_NEW);
            } catch (RemoteException e1) {
                e1.printStackTrace();
            }
        } else if (e.getSource() == exampleBrowserButton) {
            if (!example_browser_frame_.isVisible()) {
                example_browser_frame_.setSize(300, 650);
                example_browser_frame_.setLocation(700, 0);
                this.addInternalFrame(example_browser_frame_);
                example_browser_frame_.show();
            }
        } else if (e.getSource() == methodBrowserButton) {
            if (!method_browser_frame_.isVisible()) {
                method_browser_frame_.setSize(300, 650);
                method_browser_frame_.setLocation(700, 0);
                this.addInternalFrame(method_browser_frame_);
                method_browser_frame_.show();
            }
        } else if (e.getSource() == theoryBrowserButton) {
            JOptionPane.showMessageDialog(this, "not yet implemented");
            // } else if (e.getSource() == nextStepButton) {
            // getActiveWorksheet().nextStep();
            // } else if (e.getSource() == autocalcButton) {
            // getActiveWorksheet().autoCalc();
        } else if (e.getSource() == tacticsButton) {
            JOptionPane.showMessageDialog(this, "not yet implemented");
        } else { // invalid source
            logger.error("invalid source for ActionEvent");
        }
        // Formalization f = new Formalization();
        // openNewWorksheet(f, WorksheetDialog.STARTFROM_EMPTY_WORKSHEET,
        // CalcHead.VIEWSTYLE_SIMPLE_SOLVE);
    }

    /**
     * to set the problem browser visible
     */
    private void setProblemBrowserVisible() {
        addInternalFrame(problem_browser_frame_);
        problem_browser_frame_.setSize(500, 650);
        problem_browser_frame_.setLocation(500, 0);
        problem_browser_frame_.show();
    }

    /**
     * creates a new problem browser frame, inserts it to the internal frames
     * and registeres it to the problem browser dialog
     */
    private void generateProblemBrowser() {
        problem_browser_frame_ = new BrowserFrame("Problem Browser",
                new ProblemBrowserPanel(this, loadPblHierarchy()));
        try {
            BDialog problem_browser_dialog = InformationProcessor.getSDialog()
                    .getProblemBrowserDialog(session_id_);
            BrowserFrameRMI browser_frame_rmi = new BrowserFrameRMI(
                    problem_browser_frame_);

            problem_browser_dialog.registerBrowserFrame(browser_frame_rmi);
            problem_browser_frame_
                    .registerBrowserDialog(problem_browser_dialog);

        } catch (RemoteException remote_exception) {
            remote_exception.printStackTrace();
        }
    }

    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                WindowApplication windowApplication = new WindowApplication();
                if (logger.isDebugEnabled())
                    logger.debug("WindowApplication started");
            }
        });
    }

    /*
     * (non-Javadoc)
     * 
     * @see isac.gui.ILoginListener#loginEventOccured(isac.gui.LoginEvent)
     */
    public void loginEventOccured(LoginEvent loginEvent) {
        int type = loginEvent.getType();
        switch (type) {
        case (LoginEvent.LOGIN_TYPE_TRY_TO_LOGIN):
            String username = loginEvent.getUsername();
            String password = loginEvent.getPassword();
            this.login(username, password);
            break;
        case (LoginEvent.LOGIN_TYPE_EXIT):
            this.dispose();
            break;
        }
    }

}