/****************************************************************** 
 * Copyright (c) 2003, Mario Hochreiter, Student of 
 * Software Engineering at FH-Hagenberg, Austria. 
 * Project member of the isac team at the Institute for 
 * Software-Technologie, Graz University of Technology, Austria. 
 * 
 * Use is subject to license terms.
 ******************************************************************/
package isac.gui.treetable;

import isac.interfaces.ICalcElement;
import isac.interfaces.ICalcIterator;
import isac.util.CalcChangedEvent;

import java.io.Serializable;
import java.rmi.RemoteException;

import org.apache.log4j.Logger;

/**
 * 
 * @author Mario Hochreiter
 * @version 0.1 18/09/2003
 * @version 1.1
 * @author Alois Kirchsteiger 2005
 */
public class CalculationModel extends AbstractTreeModel implements Serializable {
	static Logger logger = Logger.getLogger(CalculationModel.class.getName());

	protected Object root;

	protected CalcModelHierarchy hierarchy_;

	protected ICalcIterator it_;

	/**
	 * @param root
	 *            the root element of the <code>CalculationModel</code>
	 */
	public CalculationModel(CalcModelHierarchy h, ICalcIterator it) {
		super(h.getRoot());
		root = h.getRoot();
		it_ = it;
		hierarchy_ = h;
		try {
			/*
			 * the following three lines were created to append an empty node at
			 * the end of the worksheet
			 * 
			 * @see isac.gui.util.CalcModelHierarchy
			 */
			Object[] oa = hierarchy_.addInsertionNode();
			fireTreeNodesInserted(this, (Object[]) oa[0], (int[]) oa[1],
					(Object[]) oa[2]);
			fireTreeStructureChanged(this, (Object[]) oa[0], (int[]) oa[1],
					(Object[]) oa[2]);
			it_.moveRoot();
			//insert(it_); LK050728
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @return The root element of the <code>CalculationTree</code>
	 * @throws IllegalArgumenException
	 */
	public Object getRoot() {
		if (root instanceof CalcModelNode) {
			return root;
		}
		throw new IllegalArgumentException(
				"CalculationModel can only handle objects"
						+ "of type CalcModelHierarchy");
	}

	/**
	 * @see javax.swing.tree.TreeModel#getChildCount(java.lang.Object)
	 */
	public int getChildCount(Object parent) {
		return hierarchy_.getChildCount((CalcModelNode) parent);
	}

	/**
	 * @see javax.swing.tree.TreeModel#getChild(java.lang.Object, int)
	 */
	public Object getChild(Object parent, int index) {
		return ((CalcModelNode) parent).getChild(index);
		//    return hierarchy_.getChild((CalcModelNode)parent, index);
	}

	public void setCalcModelHierarchy(CalcModelHierarchy h) {
		this.hierarchy_ = h;
	}


	/**
	 * resets the tree to nothing but the root element
	 */
	private void reset() {
		fireTreeNodesRemoved(this, new Object[] {}, new int[] { 0 },
				new Object[] { hierarchy_.getRoot() });
		hierarchy_.reset();
	}

	/**
	 * rebuilds the whole tree from the start
	 */
	public void rebuild() {
		try {
			reset();
			it_.moveRoot();
			insert(it_);
			nSteps();
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

	/**
	 * inserts new nodes as long as the iterator can move down
	 */
	public void nSteps() {
		try {
			while (it_.moveDown()) {
				insert(it_);
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

	/**
	 * updates the tree from a calcChangedEvent
	 * 
	 * WN0504DI the ONLY call in <code>Worksheet#calcChanged</code> assumes
	 * that ALL formulae between
	 * <code>CalcChangedEvent#getLastUnchangedFormula</code> and
	 * <code>CalcChangedEvent#getLastGeneratedFormula</code> are displayed on
	 * the <code>Worksheet</code>.
	 */
	public void calcChanged(CalcChangedEvent e) {
		try {
			ICalcIterator it = e.getLastUnchangedFormula();
//			something has been deleted
			if (e.getLastDeletedFormula()
					.compareTo(e.getLastUnchangedFormula()) != 0) { 
				if (false) { //TODO what kind of iterator for JTree?
					//FIXME.WN050511 this CANNOT work: moveDown references
					// Nodes already deleted
					while (it.moveDown()
							&& (it.compareTo(e.getLastDeletedFormula()) != 0)) {
						remove(it);
					}
				}
			}
			if (e.getLastGeneratedFormula() != null) {
				it = e.getLastUnchangedFormula();
				//if lastGeerated = lastUnchanged, only the lastGenerated should be inserted
				if (it.compareTo(e.getLastGeneratedFormula()) == 0) {
					// insert the lastGenerated
					insert(it);
				} else {
					while (it.moveDown()
						&& it.compareTo(e.getLastGeneratedFormula()) != 0) {
						insert(it);
					}
					insert(it);
				}
			}
		} catch (RemoteException ex) {
			ex.printStackTrace();
		}
	}

	/**
	 * informs the tree to insert one node that is specified by a Object[] oa
	 * where oa[0] = path to the parent oa[1] = int[1] with the index of the
	 * inserted node as only element oa[2] = Object[1] with the inserted node as
	 * only element
	 * 
	 * @author Alois Kirchsteiger 09.12.2004 13:55:24
	 */
	private void insert(ICalcIterator it) {
		Object[] oa = hierarchy_.insert((ICalcIterator) it.clone());
		fireTreeNodesInserted(this, (Object[]) oa[0], (int[]) oa[1],
				(Object[]) oa[2]);
		fireTreeStructureChanged(this, (Object[]) oa[0], (int[]) oa[1],
				(Object[]) oa[2]);
	}

	/**
	 * informs the tree to remove one node that is specified by a Object[] oa
	 * where oa[0] = path to the parent oa[1] = int[1] with the index of the
	 * deleted node as only element oa[2] = Object[1] with the deleted node as
	 * only element
	 * 
	 * @author Alois Kirchsteiger 09.12.2004 13:55:24
	 */
	private void remove(ICalcIterator it) {
		Object[] oa = hierarchy_.delete((ICalcIterator) it.clone());
		fireTreeNodesRemoved(this, (Object[]) oa[0], (int[]) oa[1],
				(Object[]) oa[2]);
		fireTreeStructureChanged(this, (Object[]) oa[0], (int[]) oa[1],
				(Object[]) oa[2]);
	}

}